////////////////////////////////////////////////////
// Code Writen by J.J.M. Lut
// Date: 2/6/2021
// Purpose: The code used by the Messaging Lantern. 
// Released under the Atribution-Noncommercial licence.
// Published on https://www.instructables.com/
////////////////////////////////////////////////////

//Library Dependency : 
// SSD1306Ascii by Bill Greiman https://github.com/greiman/SSD1306Ascii
// Arduino Json: https://github.com/bblanchon/ArduinoJson
// ESP8266: https://github.com/esp8266/Arduino
// Arduino Menu: https://github.com/neu-rah/ArduinoMenu
// click encoder: https://github.com/0xPIT/encoder/blob/master/ClickEncoder.h
// http client: https://github.com/arduino-libraries/ArduinoHttpClient
// stream flow master: https://github.com/scotts/streamflow/blob/master/streamflow.h
// Timer one: https://github.com/PaulStoffregen/TimerOne/blob/master/README.md
// u8g2: https://github.com/olikraus/u8g2
// websockets: https://github.com/Links2004/arduinoWebSockets
// adafruit: https://github.com/adafruit/Adafruit_NeoPixel
// AceButton: https://github.com/bxparks/AceButton

//The inclusion of some of the libraries
#include <Arduino.h> //standart arduino
#include <Ticker.h> //interupts
//esp 8266
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <ESP8266HTTPClient.h>
#include <WiFiClient.h>
#include <ArduinoJson.h>
//arduino menu
#include "SSD1306Ascii.h"
#include "SSD1306AsciiWire.h"
#include <menu.h>
#include <menuIO/SSD1306AsciiOut.h>
#include <menuIO/serialIO.h>
//enable this include if using esp8266
#include <menuIO/esp8266Out.h>
//rotary encoder
#include <ClickEncoder.h>
#include <menuIO/clickEncoderIn.h>
#include <menuIO/keyIn.h>
#include <menuIO/chainStream.h>
#include <menuIO/serialOut.h>
#include <menuIO/serialIn.h>
#include <Adafruit_NeoPixel.h>//NEOpixel
//////////////////////////////////////////////////////////////
////////////////////////Settings Please adjust to your liking
//////////////////////////////////////////////////////////////
String USER = "USER1";      //Set to USER1 or USER2 
bool wiFi = 1;              //Default setting to turn of and off the wifi.
bool screen = 1;            //Default setting to turn on and of the Screen when not in Menu
bool touchSensor = 1;       //Default setting to enable or disable the touch sensor

//Wifi Name And Password Keep to your self and clear if sharing the code:
const char* S_S_I_D_pt = "NAME"; //Wifi network name
const char* WiFi_PW_pt = "PW"; //Wifi network password

//URLs to call when sending and recieving messages:
//URL's are in the form of "http://USER.wixsite.com/user/_functions/FUNCTION_NAME/ARGUMENT"
String getUser1DisplayMessage = "https://username.wixsite.com/mysite/_functions/getUser1DisplayMessage/USER_1";
String getUser2DisplayMessage = "https://username.wixsite.com/mysite/_functions/getUser2DisplayMessage/USER_2";
String setUser1DisplayMessage = "https://username.wixsite.com/mysite/_functions/setUser1DisplayMessage?message=";
String setUser2DisplayMessage = "https://username.wixsite.com/mysite/_functions/setUser2DisplayMessage?message=";

//Define your font here. Default font: lcd5x7
#define menuFont X11fixed7x14
#define fontW 7 //width
#define fontH 15 //height

//NEO pixel LED's library settings
//Set default startup colour and brightnes of the Messaging Lantern
int ledBrightnes = 5; //1-100 tested for, however can go up to 254 make sure supply can handle current flow!
int colourRed = 254;//0-254 
int colourBlue = 254;//0-254
int colourGreen = 254; //0-254
bool lightStatus = true; //default LED status on startup (on = true, off = false)

//Wifi setup
ESP8266WiFiMulti WiFiMulti;

/////////////////////////////////////////touch sensor
boolean touchState = HIGH;
Ticker touchTicker; //interupt timer
const int touchAnalogueRead = A0;  // ESP8266 Analog Pin ADC0 = A0
int sensorValue = 0;  // value read from the pot
int outputValue = 0;  // value to output to a PWM pin
int readTouchInterval = 3000;//read every 3 seconds
// Tracks the time since last event fired
unsigned long previousTouchMillis=0;
unsigned long currentTouchMillis = millis();


//The url information:
String HTTPUrl = "";      //URL to send to
String HTTPReadURL = "";  //URL to recieve from

String message = "Test";               //send message
String recievedMessage = "NothingYet"; //Recieved message
String payload;

///////////////////////Json buffer of size way to big.
StaticJsonDocument<2000> doc;

//////////////////reading data timing setup
int readInterval=61000;//read every 61 seconds adjust of need be
// Tracks the time since last event fired
unsigned long previousMillis=0;
unsigned long currentMillis = millis();

///////////////////////Oled setup
int screenTimeOutTime = 30;
constexpr int OLED_SDA=4;
constexpr int OLED_SDC=5;
#define I2C_ADDRESS 0x3C
SSD1306AsciiWire oled;

//////////////////////////Menu code
using namespace Menu;

/////////////////////////////////////// Encoder 
Ticker ticker;
#define encA 12
#define encB 13
//this encoder has a button here
#define encBtn 14
int buttonState=0;

///////////////////////////////////////DelayTimer made using example code: "fullidle"
//some util functions
//time ticks as boolean
template<int step>
struct Tick {
  inline operator bool() {
    return millis()>=next?next+=step,true:false;
  }
  inline void reset() {next=millis()+step;}
protected:
  unsigned long next=0;
};

////////////////////////////////////FPS as boolean // made using example code: "fullidle"
// this is an FPS limiter
//return `true` `fps` times per second (max)
//note: this can skip a frame at millis wrap (every 50days ;))
template<int fps>
struct FPS:public Tick<(1000/fps)> {};
//-----------------------------------------------------
FPS<25> menuFPS;//limit menu pool to 25 frames per second
Tick<10000> idleTimeout;


////////////////
//Menu info see MenuInit
//String outText = ""; -> message defined in url definition
bool textUpdate = false;

////////////////////////////////////setup of the encoder
//macro to create navigation control root object (nav) using mainMenu
serialIn serial(Serial);
ClickEncoder clickEncoder(encA,encB,encBtn,2);
ClickEncoderStream encStream(clickEncoder,1);
MENU_INPUTS(in,&encStream,&serial);
void timerIsr() {clickEncoder.service();}////////////interupt event


///////////////////////////////////////////////////////// neopixel led strip setup:
// On a Trinket or Gemma we suggest changing this to 1:
#define LED_PIN    2
// How many NeoPixels are attached to the ESP in a single collum?
#define LED_COUNT 16

//Fixed colour mode
bool FixedColour;

// Declare our NeoPixel strip object:
Adafruit_NeoPixel strip(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ800);

///////////////////
// MENU init See example code of Arduino Menu library for explaination to what everything does
//////////////////
result doAlert(eventMask e, prompt &item);
result zZz() {Serial.println("zZz");return proceed;}
result showEvent(eventMask e,navNode& nav,prompt& item) {
  Serial.print("event: ");
  Serial.println(e);
  return proceed;
}
float test=55;

result action1(eventMask e) {
  Serial.print(e);
  Serial.println(" action1 executed, proceed menu");
  Serial.flush();
  return proceed;
}

result action2(eventMask e,navNode& nav,prompt& item) {
  Serial.print(e);
  Serial.print(" action2 executed, quiting menu");
  return quit;
}

int ledCtrl=LOW;

result myLedOn() {
  ledCtrl=HIGH;
  Serial.println("LED-HIGH");
  return proceed;
}
result myLedOff() {
  ledCtrl=LOW;
  Serial.println("LED-LOW");
  return proceed;
}
//Navigation menu's
MENU(subMenu0,"Sweet!",showEvent,noEvent,wrapStyle
  ,OP("you're cute!",subMenu0Text0,enterEvent) //A selection shows: Text, Requires a callback func , type of event for selection of menu option.
  ,OP("Love You",subMenu0Text1,enterEvent)
  ,OP("You can do it!",subMenu0Text2,enterEvent)
  ,OP("My Hero",subMenu0Text3,enterEvent)
  ,EXIT("<Back")
);

MENU(subMenu1,"Attention!",showEvent,noEvent,wrapStyle
  ,OP("Look Insta!",subMenu1Text0,enterEvent)
  ,OP("SnapStreak!",subMenu1Text1,enterEvent)
  ,OP("You there?",subMenu1Text2,enterEvent)
  ,OP("Messenger Now!",subMenu1Text3,enterEvent)
  ,OP("Babe!?!?",subMenu1Text4,enterEvent)
  ,EXIT("<Back")
);

MENU(subMenu2,"General",showEvent,noEvent,wrapStyle
  ,OP("Hi from Amsterdam!",subMenu2Text0,enterEvent)
  ,OP("Wanna Call?",subMenu2Text1,enterEvent)
  ,OP("Amazing!",subMenu2Text2,enterEvent)
  ,OP("Im cold..",subMenu2Text3,enterEvent)
  ,EXIT("<Back")
);

MENU(subMenu3,"Example",showEvent,noEvent,wrapStyle
  ,OP("The Circuit Works",subMenu3Text0,enterEvent)
  ,OP("Instructables",subMenu3Text1,enterEvent)
  ,OP("Messaging Lantern",subMenu3Text2,enterEvent)
  ,OP("Personal Message",subMenu3Text3,enterEvent)
  ,EXIT("<Back")
);

//Setting menu
MENU(subMenu4,"Colours",showEvent,noEvent,wrapStyle
  ,FIELD(colourRed,"Red","%",0,254,10,1,doNothing,noEvent,wrapStyle)
  ,FIELD(colourGreen,"Green","%",0,254,10,1,doNothing,noEvent,wrapStyle)
  ,FIELD(colourBlue,"Blue","%",0,254,10,1,doNothing,noEvent,wrapStyle)
  ,FIELD(ledBrightnes,"brightness","%",0,100,10,1,doNothing,noEvent,wrapStyle)
  ,OP("LED On/Off",ledToggle,enterEvent)
  ,EXIT("<Back")
);

//Settings menu
 MENU(subMenu5,"Settings",showEvent,noEvent,wrapStyle
  ,OP("wiFi On/Off",setWifi,enterEvent)
  ,OP("Screen On/Off",setScreen,enterEvent)
  ,OP("touch On/Off",setTouchOff,enterEvent)
  ,EXIT("<Back")
);

MENU(mainMenu,"Message Type",zZz,noEvent,wrapStyle
  ,OP("Category",showEvent,enterEvent)
  ,SUBMENU(subMenu2)//general
  ,SUBMENU(subMenu0)//sweet
  ,SUBMENU(subMenu1)//attention
  ,SUBMENU(subMenu3)//Example
  ,SUBMENU(subMenu4)//Colours
);

#define MAX_DEPTH 2

//define output device
idx_t serialTops[MAX_DEPTH] = {0};
serialOut outSerial(Serial, serialTops);

//describing a menu output device without macros
//define at least one panel for menu output
const panel panels[] MEMMODE = {{0, 0, 128 / fontW, 64 / fontH}};
navNode* nodes[sizeof(panels) / sizeof(panel)]; //navNodes to store navigation status
panelsList pList(panels, nodes, 1); //a list of panels and nodes
idx_t tops[MAX_DEPTH] = {0, 0}; //store cursor positions for each level
SSD1306AsciiOut outOLED(&oled, tops, pList, 8, 1+((fontH-1)>>3) ); //oled output device menu driver
menuOut* constMEM outputs[] MEMMODE = {&outOLED, &outSerial}; //list of output devices
outputsList out(outputs, sizeof(outputs) / sizeof(menuOut*)); //outputs list

//macro to create navigation control root object (nav) using mainMenu
NAVROOT(nav, mainMenu, MAX_DEPTH, in, out);

result alert(menuOut& o, idleEvent e) {
  if (e == idling) {
    o.setCursor(0, 0);
    o.print("alert test");
    o.setCursor(0, 1);
    o.print("press [select]");
    o.setCursor(0, 2);
    o.print("to continue...");
  }
  return proceed;
}

result doAlert(eventMask e, prompt &item) {
  nav.idleOn(alert);
  return proceed;
}

//when menu is suspended
result idle(menuOut &o, idleEvent e) {
  o.clear();
  if (&o==&outOLED) {
    if (e==idling) {
      o.println("Switching menu");
      o.println("Loading Messages...");
      o.println("Press to go Back.");
      delay(2000);
      oled.clear();
    }
  } else
    switch (e) {
      case idleStart: o.println("suspending menu!"); break;
      case idling: o.println("suspended..."); break;
      case idleEnd: o.println("resuming menu."); break;
    }
  return proceed;
}
//////////////
#define LEDPIN LED_BUILTIN

void setup() {
  /////////////////////serial setup
  pinMode(encBtn,INPUT_PULLUP);
  pinMode(LED_BUILTIN, OUTPUT); //turnoff buildin led
  digitalWrite(LED_BUILTIN, LOW);
  
  Serial.begin(115200);
  while (!Serial);
  Serial.println("menu 4.x test"); Serial.flush();

  ///////////////////////////////Start i2c
  Wire.begin(); //

  /////////////////////// LED init
   // These lines are specifically to support the Adafruit Trinket 5V 16 MHz.
   // Any other board, you can remove this part (but no harm leaving it):
  #if defined(__AVR_ATtiny85__) && (F_CPU == 16000000)
    clock_prescale_set(clock_div_1);
  #endif
    // END of Trinket-specific code.
  
    strip.begin();           // INITIALIZE NeoPixel strip object (REQUIRED)
    strip.setBrightness(ledBrightnes); // Set BRIGHTNESS to about 1/5 (max = 255)
    strip.clear();
    strip.show();            // Turn OFF all pixels ASAP

  /////////////////////////////////////////////// Touch sensor setup
  touchTicker.attach(0.5, touchInteruptFunc);  //analoge read of touch sensor every 0.2 seconds
  

  /////////////////////////////////////////////Wifi Setup
  WiFi.disconnect(true);
  WiFi.mode(WIFI_STA);
  WiFiMulti.addAP(S_S_I_D_pt, WiFi_PW_pt);

  ////OledSetup and splash screen
  oled.begin(&Adafruit128x64, I2C_ADDRESS);
  oled.setFont(menuFont);
  oled.clear();
  oled.setCursor(0, 0);
  oled.print("Messaging Lantern");
  oled.setCursor(0, 2);
  oled.print("Code: V2");
  oled.setCursor(0, 4);
  oled.print("By: J.J.M.Lut");
  oled.setCursor(0, 6);
  oled.print("Date: 2/6/2021");
  
  delay(3000);
  oled.clear();
  
  //Connection countdown
  for (uint8_t t = 5; t > 0; t--) {
    Serial.printf("[WiFi- SETUP] WAIT %d...\n", t);
    oled.clear(); 
    oled.print("[WiFi-SETUP] WAIT... " + String(t));
    Serial.flush();
    delay(1000);
    oled.clear(); 
  }

  ///////////////////////////Menu setup
  nav.timeOut=screenTimeOutTime;//made very short for demo purposes
  nav.idleTask = idle; //point a function to be used when menu is suspended
  mainMenu[0].enabled=disabledStatus;

  
  ////////////////////////Interupt timer setup
  ticker.attach(0.01,timerIsr);
  //Timer1.initialize(1000);// use for Arduino
  //Timer1.attachInterrupt(timerIsr); // use for arduino

//WIFI URL INFO
  if (USER == "USER1"){
    HTTPUrl = setUser2DisplayMessage; //The url to send to.
    HTTPReadURL = getUser1DisplayMessage; // 
  }
  else{
    HTTPUrl = setUser1DisplayMessage; //The url to send to.
    HTTPReadURL = getUser2DisplayMessage;

}

}

void loop() {

  
 ////////////////////////////////////MenuTimeout
 if(nav.sleepTask) {
    //do your own idle state stuff here
    //we can now use encoder, buttons, serial, display or whatever input/output we have
    // because menu is 100% dormant
     // Get snapshot of time 
     currentMillis = millis();
     
     //////////////////////////////////////////////
     ////////Read web while idle. Every pre defined interval
     /////////////////////////////////////////////
     
     // How much time has passed, accounting for rollover with subtraction!
     Serial.println("TimeDiff = " + String(currentMillis - previousMillis));
     if ((unsigned long)(currentMillis - previousMillis) >= readInterval) {
        Serial.println("Initiate Read");
        // It's time to do something!
        bool res = WifiRead(HTTPReadURL);
   
        // Use the snapshot to set track time until next event
        if (res){
          previousMillis = currentMillis;
        }
     }
    
    buttonState=digitalRead(encBtn); // put your main code here, to run repeatedly:
     if (buttonState == 1){
        //Serial.println("ButtonLow");
      }
     if (buttonState==0){
       //Serial.println("ButtonHigh");
       nav.idleOff();
       delay(1000);
     }
    //Serial.write(Serial.available()?(char)Serial.read():'.');
    delay(200);
  } 
  //////////////////////update menu corresponding to set FPS and if not idle
  else if(menuFPS) nav.poll(); /// update the menu
  
  digitalWrite(LED_BUILTIN, ledCtrl);
  
/////////////////////////////// wait for WiFi connection
  if(textUpdate == true){
    if ((WiFiMulti.run() == WL_CONNECTED)) {
      bool res =  WifiWrite(HTTPUrl, message);
      Serial.print("URL write" + String(res));
    }
    else {
        //if unable to connect to network
        Serial.printf("Wifi not connected!");
        oled.clear();
        oled.setCursor(0, 0);
        oled.println("Wifi not connected!");
        oled.setCursor(0, 2);
        oled.println("Pls. verify connection.");
        oled.setCursor(0, 2);
        delay(3000);
        textSetUpdate(false);
        oled.clear();
        nav.refresh();
      }
  }

////////////////////////update the colour
  if (lightStatus){
    colorWipe(strip.Color(colourRed,   colourGreen,   colourBlue), 1); // Red
    strip.setBrightness(ledBrightnes); // Set BRIGHTNESS to about 1/5 (max = 255)
  }else{
    Serial.println("clear led strip");
    strip.clear();
    strip.setBrightness(0); // Set BRIGHTNESS to about 1/5 (max = 255)
    strip.show(); // Update strip with new contents
  }
 
  delay(100);//simulate a delay when other tasks are done
  
}
